import os
from dotenv import load_dotenv

load_dotenv(override=True)
MODEL_NAME = os.environ.get("LITELLM_MODEL_NAME", "gpt-3.5-turbo")


def sync_test():
    from versionhq.task.model import ResponseField, Task
    from versionhq.agent.model import Agent
    from versionhq.task.model import Task, ResponseField

    agent_a = Agent(
        role="Demo Agent A",
        goal="""My amazing goals""",
        backstory="My amazing backstory",
        verbose=True,
        llm=MODEL_NAME,
        max_token=3000,
    )
    task = Task(
        description="Analyze the client's business model, target audience, and customer information and define the optimal cohort timeframe based on customer lifecycle and product usage patterns.",
        expected_output_raw=False,
        expected_output_json=True,
        output_field_list=[
            ResponseField(title="test1", type="int", required=True),
            ResponseField(title="test2", type="array", required=True),
        ],
        expected_output_pydantic=True,
        context=[],
        tools=[],
        callback=None,
    )

    print("...start to execute the task...")
    res = task.execute_sync(agent=agent_a)  # -> TaskOutput class
    print("Task is completed. ID:", res.task_id)
    print("Raw result: ", res.raw)
    print("JSON: ", res.json_dict)
    print("Pydantic: ", res.pydantic)
    return res


if __name__ == "__main__":
    sync_test()



import os
from dotenv import load_dotenv
from typing import Union, Any, Dict

from vesionhq.team.model import TeamOutput


load_dotenv(override=True)
MODEL_NAME = os.environ.get("LITELLM_MODEL_NAME", "gpt-3.5-turbo")


def kickoff_test() -> Union[TeamOutput | Dict[str, Any]]:
    from versionhq.agent.model import Agent
    from versionhq.task.model import Task, ResponseField
    from versionhq.team.model import Team, TeamMember, TaskHandlingProcess

    agent_a = Agent(
        role="Demo Agent A",
        goal="""My amazing goals""",
        backstory="My amazing backstory",
        verbose=True,
        llm=MODEL_NAME,
        max_token=3000,
    )

    agent_b1 = Agent(
        role="Demo Agent B-1",
        goal="""My amazing goals""",
        backstory="My amazing backstory",
        verbose=True,
        llm=MODEL_NAME,
        max_token=3000,
    )

    task_1 = Task(
        description="Amazing task description",
        expected_output_raw=False,
        expected_output_json=True,
        expected_output_pydantic=True,
        output_field_list=[
            ResponseField(title="field-1", type="int", required=True),
            ResponseField(
                title="field-2", type="array", required=True
            ),  # => { field-1: 3, field-2: [item-1, item2, ]}
        ],
        context=[],
        tools=[],
        callback=None,
    )

    task_2 = Task(
        description="Amazing task description",
        expected_output_raw=False,
        expected_output_json=True,
        expected_output_pydantic=True,
        output_field_list=[
            ResponseField(title="field-1", type="int", required=True),
            ResponseField(title="field-2", type="array", required=True),
        ],
        context=[],
        tools=[],
        callback=None,
    )

    team_task = Task(
        description="Amazing team task description",
        expected_output_raw=False,
        expected_output_json=True,
        expected_output_pydantic=True,
        output_field_list=[
            ResponseField(title="field-1", type="str", required=True),
        ],
    )

    team = Team(
        members=[
            TeamMember(agent=agent_a, is_manager=True, task=task_1),
            TeamMember(agent=agent_b1, is_manager=False, task=task_2),
        ],
        team_tasks=[
            team_task,
        ],
        process=TaskHandlingProcess.sequential,
        verbose=True,
        memory=False,
        before_kickoff_callbacks=[],  # add any callables
        after_kickoff_callbacks=[],
        prompt_file="sample.demo.Prompts.demo.py",
    )

    print("...team will kickoff...")
    res = team.kickoff()  # return TeamOutput
    print("Task is completed by team ID:", res.team_id)
    print("Raw result: ", res.raw)
    print("JSON: ", res.json_dict)
    print("Pydantic: ", res.pydantic)
    print("All task output: ", res.return_all_task_outputs)


if __name__ == "__main__":
    kickoff_test()



import json
import threading
import uuid
from concurrent.futures import Future
from hashlib import md5
from typing import Any, Dict, List, Set, Optional, Tuple, Callable

from pydantic import (
    UUID4,
    BaseModel,
    Field,
    PrivateAttr,
    field_validator,
    model_validator,
)
from pydantic_core import PydanticCustomError

from versionhq._utils.process_config import process_config
from versionhq.task import TaskOutputFormat
from versionhq.tool.model import Tool, ToolCalled


class ResponseField(BaseModel):
    """
    Field class to use in the response schema for the JSON response.
    """

    title: str = Field(default=None)
    type: str = Field(default=None)
    required: bool = Field(default=True)


class TaskOutput(BaseModel):
    """
    Store the final output of the task in TaskOutput class.
    Depending on the task output format, use `raw`, `pydantic`, `json_dict` accordingly.
    """

    class AgentOutput(BaseModel):
        """
        Keep adding agents' learning and recommendation and store it in `pydantic` field of `TaskOutput` class.
        Since the TaskOutput class has `agent` field, we don't add any info on the agent that handled the task.
        """

        customer_id: str = Field(
            default=None, max_length=126, description="customer uuid"
        )
        customer_analysis: str = Field(
            default=None, max_length=256, description="analysis of the customer"
        )
        business_overview: str = Field(
            default=None,
            max_length=256,
            description="analysis of the client's business",
        )
        cohort_timeframe: int = Field(
            default=None,
            max_length=256,
            description="Suitable cohort timeframe in days",
        )
        kpi_metrics: List[str] = Field(
            default=list, description="Ideal KPIs to be tracked"
        )
        assumptions: List[Dict[str, Any]] = Field(
            default=list, description="assumptions to test"
        )

    task_id: UUID4 = Field(
        default_factory=uuid.uuid4, frozen=True, description="store Task ID"
    )
    raw: str = Field(default="", description="Raw output of the task")
    pydantic: Optional[BaseModel | AgentOutput] = Field(
        default=None, description="Pydantic output of task"
    )
    json_dict: Optional[Dict[str, Any]] = Field(
        default=None, description="JSON dictionary of task"
    )

    def __str__(self) -> str:
        return (
            str(self.pydantic)
            if self.pydantic
            else str(self.json_dict) if self.json_dict else self.raw
        )

    @property
    def json(self) -> Optional[str]:
        if self.output_format != TaskOutputFormat.JSON:
            raise ValueError(
                """
                Invalid output format requested.
                If you would like to access the JSON output,
                pleae make sure to set the output_json property for the task
                """
            )
        return json.dumps(self.json_dict)

    def to_dict(self) -> Dict[str, Any]:
        """Convert json_output and pydantic_output to a dictionary."""
        output_dict = {}
        if self.json_dict:
            output_dict.update(self.json_dict)
        elif self.pydantic:
            output_dict.update(self.pydantic.model_dump())
        return output_dict


class Task(BaseModel):
    """
    Task to be executed by the agent or the team.
    Each task must have a description and at least one expected output format either Pydantic, Raw, or JSON, with necessary fields in ResponseField.
    Then output will be stored in TaskOutput class.
    """

    __hash__ = object.__hash__

    id: UUID4 = Field(
        default_factory=uuid.uuid4,
        frozen=True,
        description="unique identifier for the object, not set by user",
    )
    name: Optional[str] = Field(default=None)
    description: str = Field(description="Description of the actual task")
    _original_description: str = PrivateAttr(default=None)

    # output
    expected_output_raw: bool = Field(default=False)
    expected_output_json: bool = Field(default=True)
    expected_output_pydantic: bool = Field(default=False)
    output_field_list: Optional[List[ResponseField]] = Field(
        default=[
            ResponseField(title="output", type="str", required=True),
        ]
    )
    output: Optional[TaskOutput] = Field(
        default=None, description="store the final task output in TaskOutput class"
    )

    # task setup
    context: Optional[List["Task"]] = Field(
        default=None, description="other tasks whose outputs should be used as context"
    )
    tools_called: Optional[List[ToolCalled]] = Field(
        default_factory=list, description="tools that the agent can use for this task"
    )
    take_tool_res_as_final: bool = Field(
        default=False,
        description="when set True, tools res will be stored in the `TaskOutput`",
    )

    prompt_context: Optional[str] = None
    async_execution: bool = Field(
        default=False,
        description="whether the task should be executed asynchronously or not",
    )
    config: Optional[Dict[str, Any]] = Field(
        default=None, description="configuration for the agent"
    )
    callback: Optional[Any] = Field(
        default=None, description="callback to be executed after the task is completed."
    )

    # recording
    processed_by_agents: Set[str] = Field(default_factory=set)
    used_tools: int = 0
    tools_errors: int = 0
    delegations: int = 0

    @property
    def output_prompt(self):
        """
        Draft prompts on the output format by converting `output_field_list` to dictionary.
        """

        output_prompt, output_dict = "", dict()
        for item in self.output_field_list:
            output_dict[item.title] = f"your answer in {item.type}"

        output_prompt = f"""
        The output formats include the following format:
        {output_dict}
        """
        return output_prompt

    @property
    def expected_output_formats(self) -> List[TaskOutputFormat]:
        outputs = []
        if self.expected_output_json:
            outputs.append(TaskOutputFormat.JSON)
        if self.expected_output_pydantic:
            outputs.append(TaskOutputFormat.PYDANTIC)
        if self.expected_output_raw:
            outputs.append(TaskOutputFormat.RAW)
        return outputs

    @property
    def key(self) -> str:
        output_format = (
            TaskOutputFormat.JSON
            if self.expected_output_json == True
            else (
                TaskOutputFormat.PYDANTIC
                if self.expected_output_pydantic == True
                else TaskOutputFormat.RAW
            )
        )
        source = [self.description, output_format]
        return md5("|".join(source).encode(), usedforsecurity=False).hexdigest()

    @property
    def summary(self) -> str:
        return f"""
        Task: {self.id} - {self.description}
        "task_description": {self.description}
        "task_expected_output": {self.output_prompt}
        "task_tools": {", ".join([tool_called.tool.name for tool_called in self.tools_called])}
        """

    # validators
    @model_validator(mode="before")
    @classmethod
    def process_model_config(cls, values: Dict[str, Any]):
        return process_config(values_to_update=values, model_class=cls)

    @field_validator("id", mode="before")
    @classmethod
    def _deny_user_set_id(cls, v: Optional[UUID4]) -> None:
        if v:
            raise PydanticCustomError(
                "may_not_set_field", "This field is not to be set by the user.", {}
            )

    @model_validator(mode="after")
    def validate_required_fields(self):
        required_fields = [
            "description",
        ]
        for field in required_fields:
            if getattr(self, field) is None:
                raise ValueError(
                    f"{field} must be provided either directly or through config"
                )
        return self

    @model_validator(mode="after")
    def set_attributes_based_on_config(self) -> "Task":
        """
        Set attributes based on the agent configuration.
        """

        if self.config:
            for key, value in self.config.items():
                setattr(self, key, value)
        return self

    @model_validator(mode="after")
    def validate_output_format(self):
        if (
            self.expected_output_json == False
            and self.expected_output_pydantic == False
            and self.expeceted_output_raw == False
        ):
            raise PydanticCustomError("Need to choose at least one output format.")
        return self

    @model_validator(mode="after")
    def backup_description(self):
        if self._original_description == None:
            self._original_description = self.description
        return self

    def prompt(self, customer=str | None, product_overview=str | None) -> str:
        """
        Return the prompt of the task.
        """

        task_slices = [
            self.description,
            f"Customer overview: {customer}",
            f"Product overview: {product_overview}",
            f"Follow the output formats decribled below. Your response should NOT contain any other element from the following formats.: {self.output_prompt}",
        ]
        return "\n".join(task_slices)

    def _export_output(
        self, result: Any
    ) -> Tuple[Optional[BaseModel], Optional[Dict[str, Any]]]:
        output_pydantic: Optional[BaseModel] = None
        output_json: Optional[Dict[str, Any]] = None
        dict_output = None

        if isinstance(result, str):
            try:
                dict_output = json.loads(result)
            except json.JSONDecodeError:
                try:
                    dict_output = eval(result)
                except:
                    try:
                        import ast

                        dict_output = ast.literal_eval(result)
                    except:
                        dict_output = None

        if self.expected_output_json:
            if isinstance(result, dict):
                output_json = result
            elif isinstance(result, BaseModel):
                output_json = result.model_dump()
            else:
                output_json = dict_output

        if self.expected_output_pydantic:
            if isinstance(result, BaseModel):
                output_pydantic = result
            elif isinstance(result, dict):
                output_json = result
            else:
                output_pydantic = None

        return output_json, output_pydantic

    def _get_output_format(self) -> TaskOutputFormat:
        if self.output_json == True:
            return TaskOutputFormat.JSON
        if self.output_pydantic == True:
            return TaskOutputFormat.PYDANTIC
        return TaskOutputFormat.RAW

    def interpolate_inputs(self, inputs: Dict[str, Any]) -> None:
        """
        Interpolate inputs into the task description and expected output.
        """
        if inputs:
            self.description = self._original_description.format(**inputs)
            # self.expected_output = self._original_expected_output.format(**inputs)

    # task execution
    def execute_sync(self, agent, context: Optional[str] = None) -> TaskOutput:
        """
        Execute the task synchronously.
        """
        return self._execute_core(agent, context)

    def execute_async(self, agent, context: Optional[str] = None) -> Future[TaskOutput]:
        """
        Execute the task asynchronously.
        """

        future: Future[TaskOutput] = Future()
        threading.Thread(
            daemon=True,
            target=self._execute_task_async,
            args=(agent, context, future),
        ).start()
        return future

    def _execute_task_async(
        self, agent, context: Optional[str], future: Future[TaskOutput]
    ) -> None:
        """Execute the task asynchronously with context handling."""
        result = self._execute_core(agent, context)
        future.set_result(result)

    def _execute_core(self, agent, context: Optional[str]) -> TaskOutput:
        """
        Run the core execution logic of the task.
        """

        self.prompt_context = context
        result = agent.execute_task(task=self, context=context)
        output_json, output_pydantic = self._export_output(result)
        task_output = TaskOutput(
            task_id=self.id,
            raw=result,
            pydantic=output_pydantic,
            json_dict=output_json,
        )
        self.output = task_output
        self.processed_by_agents.add(agent.role)

        # self._set_end_execution_time(start_time)

        if self.callback:
            self.callback(self.output)

        # if self._execution_span:
        #     # self._telemetry.task_ended(self._execution_span, self, agent.team)
        #     self._execution_span = None

        # if self.output_file:
        #     content = (
        #         json_output
        #         if json_output
        #         else pydantic_output.model_dump_json() if pydantic_output else result
        #     )
        #     self._save_file(content)

        return task_output


class ConditionalTask(Task):
    """
    A task that can be conditionally executed based on the output of another task.
    Use this with `Team`.
    """

    condition: Callable[[TaskOutput], bool] = Field(
        default=None,
        description="max. number of retries for an agent to execute a task when an error occurs.",
    )

    def __init__(
        self,
        condition: Callable[[Any], bool],
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.condition = condition

    def should_execute(self, context: TaskOutput) -> bool:
        """
        Decide whether the conditional task should be executed based on the provided context.
        Return `True` if it should be executed.
        """
        return self.condition(context)

    def get_skipped_task_output(self):
        return TaskOutput(task_id=self.id, raw="", pydantic=None, json_dict=None)



import os
import uuid
from abc import ABC
from typing import Any, Dict, List, Optional, TypeVar, Union
from dotenv import load_dotenv
from pydantic import UUID4, BaseModel, Field, InstanceOf, PrivateAttr, model_validator

from versionhq._utils.cache_handler import CacheHandler
from versionhq._utils.logger import Logger
from versionhq._utils.rpm_controller import RPMController
from versionhq._utils.usage_metrics import UsageMetrics
from versionhq.agent.parser import AgentAction
from versionhq.llm.llm_vars import LLM_VARS
from versionhq.llm.model import LLM, DEFAULT_CONTEXT_WINDOW
from versionhq.task import TaskOutputFormat
from versionhq.task.model import ResponseField
from versionhq.tool.model import Tool, ToolCalled
from versionhq.tool.tool_handler import ToolHandler

load_dotenv(override=True)
T = TypeVar("T", bound="Agent")


# def _format_answer(agent, answer: str) -> Union[AgentAction, AgentFinish]:
#     return AgentParser(agent=agent).parse(answer)

# def mock_agent_ops_provider():
#     def track_agent(*args, **kwargs):
#         def noop(f):
#             return f
#         return noop
#     return track_agent

# track_agent = mock_agent_ops_provider()

# agentops = None
# if os.environ.get("AGENTOPS_API_KEY"):
#     try:
#         from agentops import track_agent
#     except ImportError:
#         track_agent = mock_agent_ops_provider()
# else:
#    track_agent = mock_agent_ops_provider()


class TokenProcess:
    total_tokens: int = 0
    prompt_tokens: int = 0
    cached_prompt_tokens: int = 0
    completion_tokens: int = 0
    successful_requests: int = 0

    def sum_prompt_tokens(self, tokens: int):
        self.prompt_tokens = self.prompt_tokens + tokens
        self.total_tokens = self.total_tokens + tokens

    def sum_completion_tokens(self, tokens: int):
        self.completion_tokens = self.completion_tokens + tokens
        self.total_tokens = self.total_tokens + tokens

    def sum_cached_prompt_tokens(self, tokens: int):
        self.cached_prompt_tokens = self.cached_prompt_tokens + tokens

    def sum_successful_requests(self, requests: int):
        self.successful_requests = self.successful_requests + requests

    def get_summary(self) -> UsageMetrics:
        return UsageMetrics(
            total_tokens=self.total_tokens,
            prompt_tokens=self.prompt_tokens,
            cached_prompt_tokens=self.cached_prompt_tokens,
            completion_tokens=self.completion_tokens,
            successful_requests=self.successful_requests,
        )


# @track_agent()
class Agent(ABC, BaseModel):
    """
    Base class for third-party agents that LLM runs.
    The agent can execute tasks alone or team. When the agent belongs to team, it needs to prioritize the team.
    * (Temp) Comment out all the optional fields except for Team and LLM settings for convenience.
    """

    __hash__ = object.__hash__
    _logger: Logger = PrivateAttr(default_factory=lambda: Logger(verbose=False))
    _rpm_controller: Optional[RPMController] = PrivateAttr(default=None)
    _request_within_rpm_limit: Any = PrivateAttr(default=None)
    _token_process: TokenProcess = PrivateAttr(default_factory=TokenProcess)
    _times_executed: int = PrivateAttr(default=0)

    id: UUID4 = Field(default_factory=uuid.uuid4, frozen=True)
    agent_ops_agent_name: str = None
    agent_ops_agent_id: str = None
    role: str = Field(description="role of the agent - used in summary and logs")
    goal: str = Field(
        description="concise goal of the agent (details are set in the Task instance)"
    )
    backstory: str = Field(description="context passed to the LLM")

    # tools
    tools: Optional[List[Any]] = Field(default_factory=list)
    tool_handler: InstanceOf[ToolHandler] = Field(
        default=None, description="handle tool cache and last used tool"
    )

    # team, rules of task executions
    team: Optional[List[Any]] = Field(
        default=None, description="Team to which the agent belongs"
    )
    allow_delegation: bool = Field(
        default=False,
        description="Enable agent to delegate and ask questions among each other",
    )
    allow_code_execution: Optional[bool] = Field(
        default=False, description="Enable code execution for the agent."
    )
    max_retry_limit: int = Field(
        default=2,
        description="max. number of retries for the task execution when an error occurs. cascaed to the `invoke` function",
    )
    max_iter: Optional[int] = Field(
        default=25,
        description="max. number of iterations for an agent to execute a task",
    )
    step_callback: Optional[Any] = Field(
        default=None,
        description="Callback to be executed after each step of the agent execution",
    )

    # llm settings cascaded to the LLM model
    llm: Union[str, InstanceOf[LLM], Any] = Field(default=None)
    function_calling_llm: Union[str, InstanceOf[LLM], Any] = Field(default=None)
    respect_context_window: bool = Field(
        default=True,
        description="Keep messages under the context window size by summarizing content",
    )
    max_tokens: Optional[int] = Field(
        default=None, description="max. number of tokens for the agent's execution"
    )
    max_execution_time: Optional[int] = Field(
        default=None, description="max. execution time for an agent to execute a task"
    )
    max_rpm: Optional[int] = Field(
        default=None,
        description="max. number of requests per minute for the agent execution",
    )

    # prompt rules
    use_system_prompt: Optional[bool] = Field(
        default=True, description="Use system prompt for the agent"
    )
    system_template: Optional[str] = Field(
        default=None, description="System format for the agent."
    )
    prompt_template: Optional[str] = Field(
        default=None, description="Prompt format for the agent."
    )
    response_template: Optional[str] = Field(
        default=None, description="Response format for the agent."
    )

    # config, cache, error handling
    config: Optional[Dict[str, Any]] = Field(
        default=None, exclude=True, description="Configuration for the agent"
    )
    cache: bool = Field(
        default=True, description="Whether the agent should use a cache for tool usage."
    )
    cache_handler: InstanceOf[CacheHandler] = Field(
        default=None, description="An instance of the CacheHandler class."
    )
    formatting_errors: int = Field(
        default=0, description="Number of formatting errors."
    )
    verbose: bool = Field(
        default=True, description="Verbose mode for the Agent Execution"
    )

    def __repr__(self):
        return f"Agent(role={self.role}, goal={self.goal}, backstory={self.backstory})"

    @model_validator(mode="after")
    def set_up_llm(self):
        """
        Set up the base model and function calling model (if any) using the LLM class.
        Pass the model config params: `llm`, `max_tokens`, `max_execution_time`, `step_callback`,`respect_context_window` to the LLM class.
        The base model is selected on the client app, else use the default model.
        """

        self.agent_ops_agent_name = self.role
        unaccepted_attributes = [
            "AWS_ACCESS_KEY_ID",
            "AWS_SECRET_ACCESS_KEY",
            "AWS_REGION_NAME",
        ]
        callbacks = (
            [
                self.step_callback,
            ]
            if self.step_callback is not None
            else []
        )

        if isinstance(self.llm, LLM):
            self.llm.timeout = self.max_execution_time
            self.llm.max_token = self.max_tokens
            self.llm.context_window_size = (
                self.llm.get_context_window_size()
                if self.respect_context_window == True
                else DEFAULT_CONTEXT_WINDOW
            )
            self.llm.callbacks = callbacks

        elif isinstance(self.llm, str):
            self.llm = LLM(
                model=self.llm,
                timeout=self.max_execution_time,
                max_tokens=self.max_tokens,
                callbacks=callbacks,
            )

            context_window_size = (
                self.llm.get_context_window_size()
                if self.respect_context_window == True
                else DEFAULT_CONTEXT_WINDOW
            )
            self.llm.context_window_size = context_window_size

        elif self.llm is None:
            model_name = os.environ.get(
                "LITELLM_MODEL_NAME", os.environ.get("MODEL", "gpt-4o-mini")
            )
            llm_params = {
                "model": model_name,
                "timeout": self.max_execution_time,
                "max_tokens": self.max_tokens,
                "callbacks": callbacks,
            }
            api_base = os.environ.get(
                "OPENAI_API_BASE", os.environ.get("OPENAI_BASE_URL", None)
            )
            if api_base:
                llm_params["base_url"] = api_base

            set_provider = model_name.split("/")[0] if "/" in model_name else "openai"
            for provider, env_vars in LLM_VARS.items():
                if provider == set_provider:
                    for env_var in env_vars:
                        key_name = env_var.get("key_name")

                        if key_name and key_name not in unaccepted_attributes:
                            env_value = os.environ.get(key_name)
                            if env_value:
                                key_name = (
                                    "api_key" if "API_KEY" in key_name else key_name
                                )
                                key_name = (
                                    "api_base" if "API_BASE" in key_name else key_name
                                )
                                key_name = (
                                    "api_version"
                                    if "API_VERSION" in key_name
                                    else key_name
                                )
                                llm_params[key_name] = env_value
                        elif env_var.get("default", False):
                            for key, value in env_var.items():
                                if key not in ["prompt", "key_name", "default"]:
                                    if key in os.environ:
                                        llm_params[key] = value
            self.llm = LLM(**llm_params)
            context_window_size = (
                self.llm.get_context_window_size()
                if self.respect_context_window == True
                else DEFAULT_CONTEXT_WINDOW
            )
            self.llm.context_window_size = context_window_size

        else:
            llm_params = {
                "model": (
                    getattr(self.llm, "model_name")
                    or getattr(self.llm, "deployment_name")
                    or str(self.llm)
                ),
                "max_tokens": (
                    getattr(self.llm, "max_tokens") or self.max_tokens or 3000
                ),
                "timeout": getattr(self.llm, "timeout", self.max_execution_time),
                "callbacks": getattr(self.llm, "callbacks") or callbacks,
                "temperature": getattr(self.llm, "temperature", None),
                "logprobs": getattr(self.llm, "logprobs", None),
                "api_key": getattr(self.llm, "api_key", None),
                "base_url": getattr(self.llm, "base_url", None),
                "organization": getattr(self.llm, "organization", None),
            }
            llm_params = {
                k: v for k, v in llm_params.items() if v is not None
            }  # factor out None values
            self.llm = LLM(**llm_params)

        """
        Set up funcion_calling LLM as well. For the sake of convenience, use the same metrics as the base LLM settings.
        """
        if self.function_calling_llm:
            if isinstance(self.function_calling_llm, LLM):
                self.function_calling_llm.timeout = self.max_execution_time
                self.function_calling_llm.max_tokens = self.max_tokens
                self.function_calling_llm.callbacks = callbacks
                context_window_size = (
                    self.function_calling_llm.get_context_window_size()
                    if self.respect_context_window == True
                    else DEFAULT_CONTEXT_WINDOW
                )
                self.function_calling_llm.context_window_size = context_window_size

            elif isinstance(self.function_calling_llm, str):
                self.function_calling_llm = LLM(
                    model=self.function_calling_llm,
                    timeout=self.max_execution_time,
                    max_tokens=self.max_tokens,
                    callbacks=callbacks,
                )
                context_window_size = (
                    self.function_calling_llm.get_context_window_size()
                    if self.respect_context_window == True
                    else DEFAULT_CONTEXT_WINDOW
                )
                self.function_calling_llm.context_window_size = context_window_size

            else:
                model_name = getattr(
                    self.function_calling_llm,
                    "model_name",
                    getattr(
                        self.function_calling_llm,
                        "deployment_name",
                        str(self.function_calling_llm),
                    ),
                )
                if model_name is not None or model_name != "":
                    self.function_calling_llm = LLM(
                        model=model_name,
                        timeout=self.max_execution_time,
                        max_tokens=self.max_tokens,
                        callbacks=callbacks,
                    )
        return self

    @model_validator(mode="after")
    def set_up_tools(self):
        """
        Similar to the LLM set up, when the agent has tools, we will declare them using the Tool class.
        """

        if not self.tools:
            pass

        else:
            tools_in_class_format = []
            for tool in self.tools:
                if isinstance(tool, Tool):
                    tools_in_class_format.append(tool)
                elif isinstance(tool, str):
                    tool_to_add = Tool(name=tool)
                    tools_in_class_format.append(tool_to_add)
                else:
                    pass
            self.tools = tools_in_class_format

        return self

    def invoke(
        self,
        prompts: str,
        output_formats: List[TaskOutputFormat],
        response_fields: List[ResponseField],
        **kwargs,
    ) -> Dict[str, Any]:
        """
        Receive the system prompt in string and create formatted prompts using the system prompt and the agent's backstory.
        Then call the base model.
        When encountering errors, we try the task execution up to `self.max_retry_limit` times.
        """

        task_execution_counter = 0

        messages = []
        messages.append({"role": "user", "content": prompts})  #! REFINEME
        messages.append({"role": "assistant", "content": self.backstory})
        print("Messages sent to the model:", messages)

        callbacks = kwargs.get("callbacks", None)

        response = self.llm.call(
            messages=messages,
            output_formats=output_formats,
            field_list=response_fields,
            callbacks=callbacks,
        )
        task_execution_counter += 1
        print("Agent's #1 res: ", response)

        if (
            response is None or response == ""
        ) and task_execution_counter < self.max_retry_limit:
            while task_execution_counter <= self.max_retry_limit:
                response = self.llm.call(
                    messages=messages,
                    output_formats=output_formats,
                    field_list=response_fields,
                    callbacks=callbacks,
                )
                task_execution_counter += 1
                print(f"Agent's #{task_execution_counter} res: ", response)

        elif response is None or response == "":
            print("Received None or empty response from LLM call.")
            raise ValueError("Invalid response from LLM call - None or empty.")

        return {"output": response.output if hasattr(response, "output") else response}

    def execute_task(self, task, context: Optional[str] = None) -> str:
        """
        Execute the task and return the output in string.
        To simplify, the tools are cascaded from the `tools_called` under the `task` Task instance if any.
        When the tools are given, the agent must use them.
        The agent must consider the context to excute the task as well when it is given.
        """

        task_prompt = task.prompt()
        # if context:
        #     task_prompt = self.i18n.slice("task_with_context").format(task=task_prompt, context=context)

        tool_results = []
        if task.tools_called:
            for tool_called in task.tools_called:
                tool_result = tool_called.tool.run()
                tool_results.append(tool_result)

            if task.take_tool_res_as_final:
                return tool_results

        # if self.team and self.team._train:
        #     task_prompt = self._training_handler(task_prompt=task_prompt)
        # else:
        #     task_prompt = self._use_trained_data(task_prompt=task_prompt)

        try:
            result = self.invoke(
                prompts=task_prompt,
                output_formats=task.expected_output_formats,
                response_fields=task.output_field_list,
            )["output"]

        except Exception as e:
            self._times_executed += 1
            if self._times_executed > self.max_retry_limit:
                raise e
            result = self.execute_task(
                task, context, [tool_called.tool for tool_called in task.tools_called]
            )

        if self.max_rpm and self._rpm_controller:
            self._rpm_controller.stop_rpm_counter()

        # for tool_result in self.tools_results:
        #     if tool_result.get("result_as_answer", False):
        #         result = tool_result["result"]

        return result



import uuid
from typing import Any, Dict, List, Callable, Type, Optional, get_args, get_origin
from pydantic import (
    UUID4,
    InstanceOf,
    BaseModel,
    ConfigDict,
    Field,
    create_model,
    field_validator,
    model_validator,
)
from pydantic_core import PydanticCustomError


class ProductProvider(BaseModel):
    """
    Store the minimal client information.
    `data_pipeline` and `destinations` are for composio plug-in.
    (!REFINEME) Create an Enum list for the options.
    (!REFINEME) Create an Enum list for regions.
    """

    id: UUID4 = Field(default_factory=uuid.uuid4, frozen=True)
    name: Optional[str] = Field(default=None, description="client name")
    region: Optional[str] = Field(
        default=None, description="region of client's main business operation"
    )
    data_pipeline: Optional[List[str]] = Field(
        default=None, description="store the data pipelines that the client is using"
    )
    destinations: Optional[List[str]] = Field(
        default=None,
        description="store the destination services that the client is using",
    )

    @field_validator("id", mode="before")
    @classmethod
    def _deny_user_set_id(cls, v: Optional[UUID4]) -> None:
        if v:
            raise PydanticCustomError(
                "may_not_set_field", "This field is not to be set by the user.", {}
            )


class Product(BaseModel):
    """
    Store the product information necessary to the outbound effrots and connect it to the `ProductProvider` instance.
    """

    id: UUID4 = Field(default_factory=uuid.uuid4, frozen=True)
    name: Optional[str] = Field(default=None, description="product name")
    description: Optional[str] = Field(
        default=None,
        max_length=256,
        description="product description scraped from landing url or client input. cascade to the agent",
    )
    provider: Optional[ProductProvider] = Field(default=None)
    audience: Optional[str] = Field(default=None, description="target audience")
    usp: Optional[str] = Field(default=None)
    landing_url: Optional[str] = Field(
        default=None, description="marketing url of the product if any"
    )
    cohort_timeframe: Optional[int] = Field(
        default=30, description="ideal cohort timeframe of the product in days"
    )

    @field_validator("id", mode="before")
    @classmethod
    def _deny_user_set_id(cls, v: Optional[UUID4]) -> None:
        if v:
            raise PydanticCustomError(
                "may_not_set_field", "This field is not to be set by the user.", {}
            )
